#version 400

uniform sampler2D s_HeightMap;
uniform sampler2D s_NormalMap;

uniform mat4 u_ModelViewProj;
uniform vec3 u_CamPosition;
uniform float u_LODFactor;
uniform float u_Radius;
uniform float u_Height;

uniform vec3 u_FaceTransform1;
uniform vec3 u_FaceTransform2;
uniform vec3 u_FaceTransform3;

vec3 spherical(vec2 uv, float h) {
	mat3 t = mat3(u_FaceTransform1, u_FaceTransform2, u_FaceTransform3);
	return normalize(t * vec3(uv, 1)) * (u_Radius + h * u_Height);
}

-- vs
layout(location = 0) in vec4 vs_Position;
out vec4 tcs_WorldPos;
out vec2 tcs_TexCoord;

void main() {
	vec2 uv = vec2(vs_Position.xy) * vec2(2) + vec2(-1);
	float h = texture2D(s_HeightMap, uv * .5+.5).x;	
	tcs_WorldPos = vec4(spherical(uv, h), 1);
	tcs_TexCoord = vs_Position.xy;
}

-- tcs
layout(vertices = 4) out;

in vec4 tcs_WorldPos[];
in vec2 tcs_TexCoord[];
out vec2 tes_TexCoord[];

vec4 project(vec4 pos) {
	vec4 result = u_ModelViewProj * pos;
	result /= result.w;
	return result;
}

bool offscreen(vec4 pos) {
	return any(lessThan(pos.xy, vec2(-1.7)) || greaterThan(pos.xy, vec2(1.7)));
}

float level(float d) {
	return clamp(u_LODFactor*2000/d, 1, 64);
}

void main() {
	#define ID gl_InvocationID
	if(ID == 0)
	{
		vec4 v0 = project(tcs_WorldPos[0]);
		vec4 v1 = project(tcs_WorldPos[1]);
		vec4 v2 = project(tcs_WorldPos[2]);
		vec4 v3 = project(tcs_WorldPos[3]);
		
		if(all(bvec4(offscreen(v0),offscreen(v1), offscreen(v2),offscreen(v3))))
		{
			gl_TessLevelOuter[0] = 0;
			gl_TessLevelOuter[1] = 0;
			gl_TessLevelOuter[2] = 0;
			gl_TessLevelOuter[3] = 0;
			
			gl_TessLevelInner[0] = 0;
			gl_TessLevelInner[1] = 0;
		}
		else
		{
			float d0 = distance(u_CamPosition, tcs_WorldPos[0].xyz);
			float d1 = distance(u_CamPosition, tcs_WorldPos[1].xyz);
			float d2 = distance(u_CamPosition, tcs_WorldPos[2].xyz);
			float d3 = distance(u_CamPosition, tcs_WorldPos[3].xyz);
			
			gl_TessLevelOuter[0] = level(mix(d3, d0, .5));
			gl_TessLevelOuter[1] = level(mix(d0, d1, .5));
			gl_TessLevelOuter[2] = level(mix(d1, d2, .5));
			gl_TessLevelOuter[3] = level(mix(d2, d3, .5));
			
			float l = max(max(gl_TessLevelOuter[0], gl_TessLevelOuter[1]), max(gl_TessLevelOuter[2], gl_TessLevelOuter[3]));
			gl_TessLevelInner[0] = l;
			gl_TessLevelInner[1] = l;
		}
	}
	
	tes_TexCoord[ID] = tcs_TexCoord[ID];
}

-- tes
layout(quads, fractional_odd_spacing, ccw) in;
//layout(quads, equal_spacing, ccw) in;

in vec2 tes_TexCoord[];
out float fs_Height;
out vec3 fs_Normal;

vec2 interpolate(vec2 bl, vec2 br, vec2 tr, vec2 tl)
{
	float u = gl_TessCoord.x;
	float v = gl_TessCoord.y;
	
	vec2 b = mix(bl,br,u);
	vec2 t = mix(tl,tr,u);	
	return mix(b,t,v);
}

void main() {
	vec2 coord = interpolate(tes_TexCoord[0], tes_TexCoord[1], tes_TexCoord[2], tes_TexCoord[3]);
	vec2 uv = vec2(coord.xy) * vec2(2) + vec2(-1);
	
	fs_Height = texture2D(s_HeightMap, uv * .5+.5).x;
	fs_Normal = texture2D(s_NormalMap, uv * .5+.5).rgb;
	
	vec4 spos = vec4(spherical(uv, fs_Height), 1);
	gl_Position = u_ModelViewProj * spos;
}

-- fs
in float fs_Height;
in vec3 fs_Normal;
out vec4 frag;

void main() {
	vec4 color = vec4(0, 1, 0, 1);
	if (fs_Height < .5) {
		color = vec4(0, 0, 1, 1);
	}
	else
	if (fs_Height > .8) {
		color = vec4(1, 1, 1, 1);
	}
	
	vec3 n = fs_Normal * 2-1;
	float light = (n.r + n.g) * -0.707 * -0.866 - n.b * -0.5;
	frag = color * light;
}